#include <Servo.h>
#include "WiFi.h"
#include "ESPAsyncWebServer.h"
 
const char* ssid = "Test";
const char* password =  "Test1234";

IPAddress local_IP(192, 168, 137, 100);
IPAddress gateway(192, 168, 1, 1);
IPAddress subnet(255, 255, 0, 0);

AsyncWebServer server(80);
AsyncWebSocket ws("/test");

char receivedChar;
boolean newData = false;

Servo servos[8];

int servo_pins[8] = {13, 14, 26, 27, 19, 32, 21, 33};
int servo_offset[8] = {0, 0, 20, 0, 0, 5, 10, 0}; 
int current_motor_position[8] = {60, 60, 60, 60, 120, 120, 120, 120};

int servo_move_speed = 5;
int body_move_speed = 7;

void onWsEvent(AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len){
  if(type == WS_EVT_CONNECT){
    Serial.println("Websocket client connection received");
  } else if(type == WS_EVT_DISCONNECT){
    Serial.println("Client disconnected");
  } else if(type == WS_EVT_DATA){
    Serial.println("Data received: ");
    receivedChar = data[len-1];
    Serial.println(receivedChar);
    newData=true;
    Serial.println();
  }
}

void setup() {
  Serial.begin(115200);
  for(int i = 0; i < 8; i++){
    servos[i].attach(servo_pins[i]);
    delay(50);
  }
  reset_servos();

  if (!WiFi.config(local_IP, gateway, subnet)) {
    Serial.println("STA Failed to configure");
  }

  WiFi.begin(ssid, password);
 
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi..");
  }
 
  Serial.println(WiFi.localIP());
 
  ws.onEvent(onWsEvent);
  server.addHandler(&ws);
 
  server.begin();
}

void reset_servos(){
  move_motor(0, 60, 50);
  move_motor(1, 60, 50);
  move_motor(2, 60, 50);
  move_motor(3, 60, 50);
  move_motor(4, 120, 50);
  move_motor(5, 120, 50);
  move_motor(6, 120, 50);
  move_motor(7, 120, 50);
}

void move_motors(int end_positions[], int delay_time){
  int counters[4] = {0, 0, 0, 0};
  int difference[4];
  int start_positions[4];
  int highest_difference = 0;
  for(int motor =0; motor<4; motor++){
    start_positions[motor] = current_motor_position[motor];
    difference[motor] = end_positions[motor] - start_positions[motor];
    if(abs(difference[motor])>highest_difference){
      highest_difference = abs(difference[motor]);
    }
    current_motor_position[motor] = end_positions[motor];
  }
  for(int i = 0; i<highest_difference; i++){
    for(int motor =0; motor<4; motor++){
      if(counters[motor] <= abs(difference[motor])){
        int current_position = start_positions[motor];
        if(difference[motor]>0){
          current_position = current_position + i;
        } else{
          current_position = current_position - i;
        }
        if(motor<=1){
          current_position = 90 + (90-current_position);
        }
        servos[motor].write(current_position+servo_offset[motor]);
        counters[motor] += 1;
      }
      
    }
    delay(delay_time);
  }
}

void move_motor_delay(int motor, int new_position){
  int difference = new_position - current_motor_position[motor];
  for(int i = 0; i<abs(difference); i++){
    int current_position = current_motor_position[motor];
    if(difference>0){
      current_position = current_position + i;
    } else{
      current_position = current_position - i;
    }
    if(motor<=1 || (motor>=4 && motor<=5)){
      current_position = 90 + (90-current_position);
    }
    servos[motor].write(current_position+servo_offset[motor]);
    delay(servo_move_speed);
  }
  current_motor_position[motor] = new_position;
}

void move_motor(int motor, int new_position, int delay_time){
  current_motor_position[motor] = new_position;
  if(motor<=1 || (motor>=4 && motor<=5)){
    new_position = 90 + (90-new_position);
  }
  servos[motor].write(new_position+servo_offset[motor]); 
  delay(delay_time);
}

void move_forward(){
  Serial.println("Moving Forward");
  
  //Move right front leg forward
  move_motor_delay(4, 160);
  move_motor_delay(0, 20);
  move_motor_delay(4, 120);

  //Move back left leg forward
  move_motor_delay(7, 160);
  move_motor_delay(3, 20);
  move_motor_delay(7, 120);

  //Move all legs
  int end_pos[4] = {60, 100, 100, 60};
  move_motors(end_pos, body_move_speed);

  //Move front left leg forward
  move_motor_delay(6, 160);
  move_motor_delay(2, 60);
  move_motor_delay(6, 120);

  //Move back right leg forward
  move_motor_delay(5, 160);
  move_motor_delay(1, 60);
  move_motor_delay(5, 120);

  reset_servos();
}

void turn_right(){
  Serial.println("Turn right");
  //Move right front leg back
  move_motor_delay(4, 160);
  move_motor_delay(0, 100);
  move_motor_delay(4, 120);

  //Move back left leg forward
  move_motor_delay(7, 160);
  move_motor_delay(3, 20);
  move_motor_delay(7, 120);

  //Move all legs
  int end_pos[4] = {60, 20, 100, 60};
  move_motors(end_pos, body_move_speed);

  //Move front left leg forward
  move_motor_delay(6, 160);
  move_motor_delay(2, 60);
  move_motor_delay(6, 120);

  //Move back right leg forward
  move_motor_delay(5, 160);
  move_motor_delay(1, 60);
  move_motor_delay(5, 120);

  reset_servos();
}

void turn_left(){
  Serial.println("Turn left");
  //Move right front leg back
  move_motor_delay(4, 160);
  move_motor_delay(0, 20);
  move_motor_delay(4, 120);

  //Move back left leg forward
  move_motor_delay(7, 160);
  move_motor_delay(3, 100);
  move_motor_delay(7, 120);

  //Move all legs
  int end_pos[4] = {60, 100, 20, 60};
  move_motors(end_pos, body_move_speed);

  //Move front left leg forward
  move_motor_delay(6, 160);
  move_motor_delay(2, 60);
  move_motor_delay(6, 120);

  //Move back right leg forward
  move_motor_delay(5, 160);
  move_motor_delay(1, 60);
  move_motor_delay(5, 120);

  reset_servos();
}

void slam(){
  Serial.println("Slamming!");

  move_motor(4, 160, 50);
  move_motor(5, 160, 50);
  move_motor(6, 160, 50);
  move_motor(7, 160, 50);

  delay(1000);

  move_motor(4, 120, 50);
  move_motor(5, 120, 50);
  move_motor(6, 120, 50);
  move_motor(7, 120, 50);
}

void loop() {
  parse_new_instructions();
}

void parse_new_instructions() {
  if (newData == true) {
    if(receivedChar =='w'){
      move_forward();
    }
    if(receivedChar =='a'){
      turn_left();
    }
    if(receivedChar =='d'){
      turn_right();
    }
    if(receivedChar =='k'){
      slam();
    }
    newData = false;
  }
}
